-- Windows file operations (tested with Euphoria 2.4)
--
-- (c) Juergen Luethje <http://home.arcor.de/luethje/>
-- mbox 1.00 -- 2003, Nov 09
-- Freeware
-- License see "mbox_eng.txt"

include dll.e
include machine.e


constant
   NIL       = -1,
   ERROR     = -1,
   UNDEFINED = -2

constant    -- TYPE WIN_FILETIME
--   dwLowDateTime  = 0,  -- C_ULONG
--   dwHighDateTime = 4,  -- C_ULONG
SIZEOF_WIN_FILETIME = 8

type WinFileTime (object x)
   -- called by set_file_time_WIN32(), local_to_utc_WIN32(),
   --   set_file_time()
   if sequence(x) and (length(x) = 2)
   and atom(x[1]) and atom(x[2]) then
      return 1
   end if
   return 0
end type


atom Kernel32 Kernel32 = NIL


--========================[ set_file_time() ]=========================--

constant
   GENERIC_WRITE              = #40000000,   -- for set_file_time_WIN32()
   FILE_SHARE_READ            = #1,
   FILE_SHARE_WRITE           = #2,
   OPEN_EXISTING              = #3,
   FILE_FLAG_BACKUP_SEMANTICS = #2000000


integer IdOpenFile IdOpenFile = UNDEFINED

function open_file_WIN32 (sequence filename, atom method)
   -- open file for read/write
   -- called by set_file_time_WIN32()
   atom lpszName, handle

   if IdOpenFile = UNDEFINED then
      IdOpenFile = define_c_func(Kernel32, "CreateFileA", {C_POINTER, C_ULONG, C_ULONG,
                                            C_POINTER, C_ULONG, C_ULONG, C_ULONG}, C_LONG)
   end if

   if IdOpenFile != ERROR then
      lpszName = allocate_string(filename)
      handle = c_func(IdOpenFile, {lpszName, method, or_bits(FILE_SHARE_READ,
                                   FILE_SHARE_WRITE), NULL, OPEN_EXISTING,
                                   FILE_FLAG_BACKUP_SEMANTICS, 0})
      free(lpszName)
      return handle      -- -1 on error
   end if
   return ERROR
end function


integer IdCloseFile IdCloseFile = UNDEFINED

procedure close_file_WIN32 (atom handle)
   -- called by set_file_time_WIN32()

   if IdCloseFile = UNDEFINED then
      IdCloseFile = define_c_proc(Kernel32, "CloseHandle", {C_ULONG})
   end if
   if IdCloseFile != ERROR then
      c_proc(IdCloseFile, {handle})
   end if
end procedure


integer IdSetFtime IdSetFtime = UNDEFINED

function set_file_time_WIN32 (sequence filename, WinFileTime timeMod)
   -- in: timeMod: time when file was last modified
   --     (number of 100ns intervals since 1601-01-01, *UTC*).
   -- called by set_file_time()
   atom handle, lpTimeMod, ret

   handle = open_file_WIN32(filename, GENERIC_WRITE)
   if handle = ERROR then
      return ERROR
   end if

   if IdSetFtime = UNDEFINED then
      IdSetFtime = define_c_func(Kernel32, "SetFileTime", {C_ULONG, C_POINTER,
                                            C_POINTER, C_POINTER}, C_LONG)
   end if

   if IdSetFtime != ERROR then
      lpTimeMod = allocate(SIZEOF_WIN_FILETIME)
      poke4(lpTimeMod, timeMod)
      ret = c_func(IdSetFtime, {handle, 0, 0, lpTimeMod}) - 1
      free(lpTimeMod)
      close_file_WIN32(handle)
      return ret                -- -1 on error
   end if
   return ERROR
end function

------------------------------------------------------------------------

integer IdLocal2Utc IdLocal2Utc = UNDEFINED

function local_to_utc_WIN32 (WinFileTime LocalTime)
   -- called by time_to_file_time_WIN32()
   atom lpLocalTime, lpUTC, ret
   WinFileTime UTC

   if IdLocal2Utc = UNDEFINED then
      if Kernel32 = NIL then
         Kernel32 = open_dll("kernel32.dll")
      end if
      IdLocal2Utc = define_c_func(Kernel32, "LocalFileTimeToFileTime",
                                  {C_POINTER, C_POINTER}, C_LONG)
   end if

   if IdLocal2Utc != ERROR then
      lpLocalTime = allocate(SIZEOF_WIN_FILETIME)
      lpUTC       = allocate(SIZEOF_WIN_FILETIME)
      poke4(lpLocalTime, LocalTime)
      ret = c_func(IdLocal2Utc, {lpLocalTime, lpUTC}) - 1
      UTC = peek4u({lpUTC, 2})
      free(lpLocalTime)
      free(lpUTC)
      if ret != ERROR then
         return UTC
      end if
   end if
   return {-1,-1}              -- error
end function


constant DAYS_UNTIL_1601_01_01 = 584389

function time_to_file_time_WIN32 (sequence t)
   -- converts the given time to WIN32 filetime format
   -- in : sequence of the form {year, month, day, hour, minute, second}
   -- out: number of 100ns intervals since 1601-01-01 as type WinFileTime
   -- called by set_file_time()
   atom secs, ns100
   tDateSerial days

   secs = (t[E_HOUR]*60 + t[E_MINUTE])*60 + t[E_SECOND]
   days = date_serial(t[E_YEAR..E_DAY]) - DAYS_UNTIL_1601_01_01
   ns100 = (days*SECS_PER_DAY + secs) * 10000000
   return local_to_utc_WIN32({remainder(ns100, #100000000), floor(ns100/#100000000)})
end function


global function set_file_time (string filename, tFileTime filetime)
   -- in : filetime as sequence of the form {year, month, day, hour, minute, second}
   -- out: 0 return on success, -1 on error
   -- called by ProcessFile()
   WinFileTime t

   t = time_to_file_time_WIN32(filetime)
   return set_file_time_WIN32(filename, t)
end function


--===========================[ del_file() ]===========================--

integer IdDelFile IdDelFile = UNDEFINED

global function del_file (string filename)
   -- called by ProcessFile()
   atom lpFilename, ret

   if IdDelFile = UNDEFINED then
      if Kernel32 = NIL then
         Kernel32 = open_dll("kernel32.dll")
      end if
      IdDelFile = define_c_func(Kernel32, "DeleteFileA", {C_POINTER}, C_LONG)
   end if

   if IdDelFile != ERROR then
      lpFilename = allocate_string(filename)
      ret = c_func(IdDelFile, {lpFilename}) - 1
      free(lpFilename)
      return ret        -- -1 on error
   end if
   return ERROR
end function

